/*=============================================================================*
 *  Copyright 2004 The Apache Software Foundation
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 *=============================================================================*/
package org.apache.ws.resource;

import org.apache.ws.addressing.EndpointReference;

/**
 * Defines a basic interface through which resources are discovered, and removed. The purpose of
 * <code>ResourceHome</code> is to interact with a collection of resources of the same type. Each resource type will
 * have its own custom implementation the <code>ResourceHome</code> interface. The implementation is expected to provide
 * custom methods for creating new resources and optionally methods that act on a set of resource objects.
 */
public interface ResourceHome extends ResourceCreation, ResourceDestruction
{
    /**
     * The resource key type. The <code>ResourceKey</code> used or passed to this <code>ResourceHome</code> must have
     * match this type (corresponds to {@link ResourceKey#getValue() ResourceKey.getValue()}).
     *
     * @return the type of the key.
     */
    String getResourceKeyClassName();

    /**
     * The name of the resource key. The <code>ResourceKey</code> used or passed to this <code>ResourceHome</code> must
     * have match this name (corresponds to {@link ResourceKey#getName() ResourceKey.getName()}).
     *
     * @return the name of the key.
     */
    String getResourceKeyName();

    /**
     * Sets the classname of the Resource Impl.
     *
     * @param className The class name of the Resource Impl.
     */
    void setResourceClassName( String className );

    /**
     * Returns the Class of the Resource Impl.
     *
     * @return The Class of the Resource Impl
     */
    String getResourceClassName();

    /**
     * Sets the Class of the ResourceKey Impl
     *
     * @param className The class name of the ResourceKey impl
     */
    void setResourceKeyClassName( String className );

    /**
     * Sets the ResourceKey QName as a String value. (i.e. QName.toString()... i.e.  {http://...}ResourceId
     *
     * @param keyName The QName.toString() representation of the ResourceKey
     */
    void setResourceKeyName( String keyName );

    /**
     * Sets the Service Impl's .Class
     *
     * @param className String representation of the class
     */
    void setServiceClassName( String className );

    /**
     * Returns the name of the Service class.
     *
     * @return the name of the Service class
     */
    String getServiceClassName();

    /**
     * Sets the target namespace of the WSDL of this home's service.
     *
     * @param targetNamespace
     */
    void setWsdlTargetNamespace( String targetNamespace );

    /**
     * Returns the target namespace of the WSDL.
     *
     * @return The String of the targetnamespace
     */
    String getWsdlTargetNamespace();

    /**
     * Retrives a resource. <b>Note:</b> This function must not return null. It must return the resource object or throw
     * an exception if there is no resource with the specified key.
     *
     * @return non-null resource
     *
     * @throws ResourceUnknownException    if no resource exists with the given key
     * @throws InvalidResourceKeyException if the resource key is invalid.
     * @throws ResourceException           if any other error occurs.
     */
    Resource find( ResourceKey key )
            throws
            ResourceUnknownException,
            InvalidResourceKeyException,
            ResourceException;

    /**
     * Removes a resource. If the resource implements the {@link RemoveCallback RemoveCallback} interface, the
     * implementation must invoke the remove() operation on the resource itself.
     *
     * @throws ResourceUnknownException    if no resource exists with the given key
     * @throws InvalidResourceKeyException if the resource key is invalid.
     * @throws RemoveNotSupportedException if remove operation is not supported.
     * @throws ResourceException           if any other error occurs.
     */
    void remove( ResourceKey key )
            throws ResourceException,
            ResourceUnknownException,
            InvalidResourceKeyException,
            RemoveNotSupportedException;

    /**
     * Retrieves the <code>Resource</code> instance that is corresponds to the id contained
     * in the specified <code>ResourceContext</code>.
     *
     * @throws ResourceUnknownException    if no resource exists with the given key
     * @throws InvalidResourceKeyException if the resource key is invalid.
     * @throws ResourceException           if any other error occurs.
     */
    Resource getInstance( ResourceContext resourceContext )
            throws
            ResourceContextException,
            ResourceUnknownException,
            ResourceException;

    /**
     * Builds an EPR based on the specified endpoint address, resource key, and WS-Addressing
     * namespace URI.
     * 
     * @param endpointAddress
     * @param key
     * @param wsAddressingURI
     * @return  An EndpointReference
     */
    // TODO: should this be moved to a static utility method?
    EndpointReference getEndpointReference( String endpointAddress, ResourceKey key, String wsAddressingURI );

}
